/////////////////////////////////////////////////////
//
//    Wii Tenkey (Wii Numberpad)
//
//    by kako
//
/////////////////////////////////////////////////////


#define _WIN32_WINNT (0x0500)
#include <windows.h>
#include <process.h>
#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
#include <string.h> 
#include <io.h>
#include <mmsystem.h>

#define FUNC __declspec(dllimport) __stdcall
extern "C" HANDLE FUNC OpenHidHandle(unsigned short vendor_id, unsigned short product_id);
extern "C" void FUNC ReadReport(HANDLE handle,unsigned char *InputReport,int *len);
extern "C" void FUNC WriteReport(HANDLE handle,unsigned char *OutputReport, int *len);
extern "C" void FUNC CloseHidHandle(HANDLE handle);

HANDLE hRsDevHandle;

const USHORT VID = 0x057e; // Nintendo (?)
const USHORT PID = 0x0306; // Wii-Remote (?)

#define MAXREPORTSIZE 256
int		InputLength,OutputLength;
unsigned char	InputReport[MAXREPORTSIZE];
unsigned char	OutputReport[MAXREPORTSIZE];

unsigned char	button1,button2;
unsigned char	Ax,Ay,Az;
unsigned char	wii_remote_loop_flag;
unsigned char	wii_remote_in_use;

void Wii_Remote_Input(void);
void Wii_Remote_mode_0x31(void);
int Wii_Remote_check(void);

const unsigned char key_table[25] = {
	0,
	' ', ' ', '+', ' ','E',' ',
	'-', '*', '9', '6','3','.',
	' ', '/', '8', '5','2','0',
	' ', 'N', '7', '4','1',' ',
};

char wav_buf1[10*1024*1024];

void wav_load(void)
{
	FILE *fp;
	long size;
	fp=fopen("test1.wav","rb");
	if (fp==NULL) {
		//printf("File Not Found\n");
	} else {
		size=filelength(fileno(fp));
		fread( wav_buf1 , 1 , size , fp );
		fclose(fp);
	}
}

void wav_play1(void)
{
	PlaySound(wav_buf1,NULL,SND_MEMORY | SND_ASYNC);	
}

void Sleep2(int msec)
{
	LARGE_INTEGER f; //\ptH[}XJE^px̒lQƗp
	LARGE_INTEGER t1,t2; //\ptH[}XJE^̃JEglQƗp
	double t;
	QueryPerformanceFrequency(&f);
	QueryPerformanceCounter(&t1);
	while (1) {
		QueryPerformanceCounter(&t2);
		t = (double)t2.QuadPart - (double)t1.QuadPart;
		t /= (double)f.QuadPart;
		t *= 1000;
		if (t>(double)msec) { return; }
	}
}

void Report_0x11(unsigned char data)
{
	OutputReport[0]=0x11; // Report ID
	OutputReport[1]=data;
	WriteReport( hRsDevHandle , OutputReport , &OutputLength);
}

void Report_0x12(unsigned char data1,unsigned char data2)
{
	OutputReport[0]=0x12; // Report ID
	OutputReport[1]=data1;
	OutputReport[2]=data2;
	WriteReport( hRsDevHandle , OutputReport , &OutputLength);
}

void Report_0x13(unsigned char data)
{
	OutputReport[0]=0x13; // Report ID
	OutputReport[1]=data;
	WriteReport( hRsDevHandle , OutputReport , &OutputLength);
}

void Report_0x16_type1(unsigned char adr,unsigned char len,unsigned char d1,unsigned char d2,unsigned char d3)
{
	OutputReport[0]=0x16; // Report ID
	OutputReport[1]=0x04; // 00(EEPROM) or 0x04(other i2c device)  or 0x08(other type2)
	OutputReport[2]=adr; // I2C ADRS
	OutputReport[3]=0xFF; // ignore
	OutputReport[4]=d1;
	OutputReport[5]=len-1;
	OutputReport[6]=d2; // data2
	OutputReport[7]=d3;
	OutputReport[8]=0x00;
	OutputReport[9]=0x00;
	OutputReport[10]=0x00; OutputReport[11]=0x00; OutputReport[12]=0x00; OutputReport[13]=0x00;
	OutputReport[14]=0x00; OutputReport[15]=0x00; OutputReport[16]=0x00; OutputReport[17]=0x00;
	OutputReport[18]=0x00; OutputReport[19]=0x00; OutputReport[20]=0x00; OutputReport[21]=0x00;
	WriteReport( hRsDevHandle , OutputReport , &OutputLength);
}

void Report_0x17(unsigned char adr,unsigned char len,unsigned char d1)
{
	OutputReport[0]=0x17; // Report ID  ... Auto Read == dump(?)
	OutputReport[1]=0x04; // 00(EEPROM) or 04(other ... 1byten)
	OutputReport[2]=adr; // I2C ADRS
	OutputReport[3]=0x00; // ignore?
	OutputReport[4]=d1;  // mem adr or command
	OutputReport[5]=0x00; // length H
	OutputReport[6]=len;  // length L
	WriteReport( hRsDevHandle , OutputReport , &OutputLength);
}

void Report_0x1a(unsigned char data)
{
	OutputReport[0]=0x1a; // Report ID
	OutputReport[1]=data;
	WriteReport( hRsDevHandle , OutputReport , &OutputLength);
}

void Wii_Remote_mode_0x30(void)
{
	// Set Report Type
	Report_0x12(0x00,0x30);  // button ONLY Report Mode
}

void Wii_Remote_mode_0x31(void)
{
	// Set Report Type
	Report_0x12(0x00,0x31);  // Acc Sensor + button  Report Mode
}

void Wii_Remote_Input(void)
{
	ReadReport( hRsDevHandle , InputReport, &InputLength);
	if ( InputReport[0]==0x30 ) {
		button1=InputReport[1]; button2=InputReport[2];
	}
	if ( InputReport[0]==0x31 ) {
		button1=InputReport[1]; button2=InputReport[2];
		Ax = InputReport[3];
		Ay = InputReport[4];
		Az = InputReport[5];
	}
}

void Wii_Remote_Input_Main(void *)
{
	// wii remote input thread
	while (wii_remote_loop_flag) {
		wii_remote_in_use=1;
		Wii_Remote_Input();
		Sleep(100);
	}
	wii_remote_in_use=0;
	return;
}

uintptr_t wii_th;

int Wii_Remote_check(void)
{
	Ax=0;
	wii_remote_loop_flag=1;
	wii_th = _beginthread( Wii_Remote_Input_Main , 0 , NULL );
	Sleep2(100);
	wii_remote_loop_flag=0;
	return (Ax!=0);
}

void Stop_Wii_remote_th(void)
{
	if (wii_remote_in_use==1) {
		TerminateThread( (HANDLE)wii_th , -1); // kill thread  if it stopped
	}
}


void SendKey(WORD ky)
{
	INPUT Key1;
	
	Key1.type           = INPUT_KEYBOARD;
	Key1.ki.wVk         = ky;
	Key1.ki.wScan       = (WORD)MapVirtualKey(ky, 0 ); // get scan code
	Key1.ki.dwFlags     = 0;
	Key1.ki.dwExtraInfo = 0;
	Key1.ki.time        = 0;
	SendInput( 1, &Key1, sizeof(INPUT) );
	Key1.ki.dwFlags     = KEYEVENTF_KEYUP;
	SendInput( 1, &Key1, sizeof(INPUT) );
}

void KeyTransfer(WORD k)
{
	if (k=='0') { SendKey('0'); }
	if (k=='1') { SendKey('1'); }
	if (k=='2') { SendKey('2'); }
	if (k=='3') { SendKey('3'); }
	if (k=='4') { SendKey('4'); }
	if (k=='5') { SendKey('5'); }
	if (k=='6') { SendKey('6'); }
	if (k=='7') { SendKey('7'); }
	if (k=='8') { SendKey('8'); }
	if (k=='9') { SendKey('9'); }
	if (k=='.') { SendKey(VK_DECIMAL); }
	if (k=='+') { SendKey(VK_ADD); }
	if (k=='-') { SendKey(VK_SUBTRACT); }
	if (k=='*') { SendKey(VK_MULTIPLY); }
	if (k=='/') { SendKey(VK_DIVIDE); }
	if (k=='E') { SendKey(VK_RETURN); }
	//if (k=='N') { SendKey(VK_ESCAPE); }
	if (k=='N') { SendKey(VK_DELETE); }
}

#define WT 1

void TenKey_read(unsigned char d , unsigned char *r1 , unsigned char *r2)
{
	int d1,d2,d3,d4,d5;
	Report_0x16_type1(0xE8, 2, 0x02 /* Data-write */, d ,0x00 ); Sleep2(WT);
	Report_0x16_type1(0xE8, 2, 0x02 /* Data-write */, d ,0x00 ); Sleep2(WT);
	Report_0x17(0xE8, 2, 0x00 /* Data-read */);
	while (1) {
		ReadReport( hRsDevHandle , InputReport, &InputLength);
		if (InputReport[0]==0x21) { break; }
		if ( ( InputReport[0]==0x30 ) || ( InputReport[0]==0x31 ) ) {
			button1=InputReport[1]; button2=InputReport[2];
		}
	}
	d1=InputReport[3]; // result ssss-ffff  ... ssss receive-size / ffff==0x07 error
	if ((d1&0x0F)==0x07) {
		//error
		d4=0xFF; d5=0xFF;
	} else {
		d2=InputReport[4]; d3=InputReport[5];  // data-offset
		d4=InputReport[6]; d5=InputReport[7]; // data
	}
	*r1 = d4;
	*r2 = d5;
}


void MainLoop(void)
{
	unsigned char d1,d2,d3;
	unsigned char key_st[4];
	int i,j,keyin,tenkey,prev;
	prev=0;
	while (1) {
		if (kbhit()) { if (getch()==0x1B) { break; } } // ESC key  to Quit
		d1=0xFC; TenKey_read(d1, &d2, &d3); key_st[0]=d3;
		d1=0xFA; TenKey_read(d1, &d2, &d3); key_st[1]=d3;
		d1=0xF6; TenKey_read(d1, &d2, &d3); key_st[2]=d3;
		d1=0xEE; TenKey_read(d1, &d2, &d3); key_st[3]=d3;
		//if (button2==0x80) { break; }  // Home Key  to Quit
		keyin= 0;
		for (i=0;i<4;i++) {
			for (j=0;j<6;j++) {
				if ( ( key_st[i] & (1<<j) )==0 ) { keyin=i*6+j+1; }
			}
		}
		tenkey=key_table[keyin];
		if (prev!=tenkey) {
			if (tenkey !=0) {
				printf("press (%c)\n",key_table[keyin]);
				KeyTransfer( key_table[keyin] );
				wav_play1();
			}
			if (prev!=0) {
				//printf("release (%c)\n",prev);
			}
		}
		prev=tenkey;
		//printf("%02X %02X %02X %02X \n",key_st[0], key_st[1], key_st[2], key_st[3] );
	}
}


int main( int argc, char *argv[])
{ 
	unsigned char d1,d2,d3;
	int peripheral_found_flag;
	
	wav_load();
	
	hRsDevHandle = OpenHidHandle(VID,PID);
	if ( hRsDevHandle == INVALID_HANDLE_VALUE) {
		MessageBox(0,"Wii-Remote not found.",0,0);
		return 1;
	}
	
	Wii_Remote_mode_0x31();
	if ( Wii_Remote_check() ) {
		
		printf("###########################################\n");
		printf("#                                         #\n");
		printf("#   Wii-Tenkey Test version 0.1           #\n");
		printf("#                  build 2007.2.11        #\n");
		printf("#                                         #\n");
		printf("###########################################\n");
		printf("\n");
		printf("Usage:\n");
		printf("   Home button  ... Quit\n");
		printf("\n");
		
		Wii_Remote_mode_0x30();
		Report_0x11(0xF0); // LEDs on
		
		peripheral_found_flag = 0;
		for (int t=0;t<79;t++) {
			Report_0x16_type1(0xE8, 2, 0x06 /* In-Out */ , 0x00 ,0xFF ); Sleep2(WT);
			d1=0xEE; TenKey_read(d1, &d2, &d3);
			if (d1==d2) {  peripheral_found_flag = 1;  break; }
			printf(".");
		}
		printf("\n");
		
		if (peripheral_found_flag==1) {
			MainLoop();
		} else {
			MessageBox(0,"Tenkey is not found.",0,0);
		}
		Report_0x11(0x10); // LEDs on
		Report_0x16_type1(0xE8, 2, 0x06 /* In-Out */ , 0xFF ,0xFF ); Sleep2(WT);
		Wii_Remote_mode_0x30();
		
		
	} else {
		
		MessageBox(0,"Wii-Remote cannot open.",0,0);
		Stop_Wii_remote_th();
	}
	CloseHidHandle( hRsDevHandle );
	return 0;
} 
